@@ -11,25 +11,14 @@ import os
 import re
 import subprocess
 import sys
-import tempfile
-import warnings
 
 import pydot
+import pydot.dot_parser
+from pydot._vendor import tempfile
 
 _logger = logging.getLogger(__name__)
 _logger.debug("pydot core module initializing")
 
-
-try:
-    from pydot import dot_parser
-except Exception as e:
-    warnings.warn(
-        "`pydot` could not import `dot_parser`, "
-        "so `pydot` will be unable to parse DOT files. "
-        f"The error was:  {e}"
-    )
-
-
 # fmt: off
 GRAPH_ATTRIBUTES = {
     "Damping", "K", "URL", "aspect", "bb", "bgcolor",
@@ -304,13 +293,13 @@ def make_quoted(s):
 
 dot_keywords = ["graph", "subgraph", "digraph", "node", "edge", "strict"]
 
-re_all_numeric = re.compile(r"^[0-9\.]+$")
+re_numeric = re.compile(r"^([0-9]+\.?[0-9]*|[0-9]*\.[0-9]+)$")
 re_dbl_quoted = re.compile(r'^".*"$', re.S)
 re_html = re.compile(r"^<.*>$", re.S)
 
-id_re_alpha_nums = re.compile(r"^[_a-zA-Z][a-zA-Z0-9_\.]*$")
+id_re_alpha_nums = re.compile(r"^[_a-zA-Z][a-zA-Z0-9_]*$")
 id_re_alpha_nums_with_ports = re.compile(
-    r'^[_a-zA-Z][a-zA-Z0-9_\.:"]*[a-zA-Z0-9_\."]+$'
+    r'^[_a-zA-Z][a-zA-Z0-9_:"]*[a-zA-Z0-9_"]+$'
 )
 id_re_with_port = re.compile(r"^([^:]*):([^:]*)$")
 
@@ -320,14 +309,20 @@ def any_needs_quotes(s):
 
     Returns True, False, or None if the result is indeterminate.
     """
-    if s.isalnum():
+
+    # Strings consisting _only_ of digits are safe unquoted
+    if s.isdigit():
         return False
 
+    # MIXED-aphanumeric values need quoting if they *start* with a digit
+    if s.isalnum():
+        return s[0].isdigit()
+
     has_high_chars = any(ord(c) > 0x7F or ord(c) == 0 for c in s)
     if has_high_chars and not re_dbl_quoted.match(s) and not re_html.match(s):
         return True
 
-    for test_re in [re_all_numeric, re_dbl_quoted, re_html]:
+    for test_re in [re_numeric, re_dbl_quoted, re_html]:
         if test_re.match(s):
             return False
 
@@ -350,7 +345,7 @@ def id_needs_quotes(s):
     # would use a reserved keyword as name. This function will return
     # false indicating that a keyword string, if provided as-is, won't
     # need quotes.
-    if s in dot_keywords:
+    if s.lower() in dot_keywords:
         return False
 
     any_result = any_needs_quotes(s)
@@ -403,7 +398,7 @@ def quote_attr_if_necessary(s):
     if not isinstance(s, str):
         return s
 
-    if s in dot_keywords:
+    if s.lower() in dot_keywords:
         return make_quoted(s)
 
     any_result = any_needs_quotes(s)
@@ -416,18 +411,24 @@ def quote_attr_if_necessary(s):
 def graph_from_dot_data(s):
     """Load graphs from DOT description in string `s`.
 
+    This function is NOT thread-safe due to the internal use of `pyparsing`.
+    Use a lock if needed.
+
     @param s: string in [DOT language](
         https://en.wikipedia.org/wiki/DOT_(graph_description_language))
 
     @return: Graphs that result from parsing.
     @rtype: `list` of `pydot.Dot`
     """
-    return dot_parser.parse_dot_data(s)
+    return pydot.dot_parser.parse_dot_data(s)
 
 
 def graph_from_dot_file(path, encoding=None):
     """Load graphs from DOT file at `path`.
 
+    This function is NOT thread-safe due to the internal use of `pyparsing`.
+    Use a lock if needed.
+
     @param path: to DOT file
     @param encoding: as passed to `io.open`.
         For example, `'utf-8'`.
@@ -660,6 +661,14 @@ class Common:
         """Get sequence"""
         return self.obj_dict["sequence"]
 
+    @staticmethod
+    def get_indent(indent, indent_level):
+        if isinstance(indent, (int, float)):
+            indent_str = " " * int(indent)
+        else:
+            indent_str = str(indent)
+        return indent_str * indent_level
+
     @staticmethod
     def _format_attr(key: str, value):
         """Turn a key-value pair into an attribute, properly quoted."""
@@ -718,7 +727,6 @@ class Node(Common):
             self.obj_dict["attributes"] = dict(attrs)
             self.obj_dict["type"] = "node"
             self.obj_dict["parent_graph"] = None
-            self.obj_dict["parent_node_list"] = None
             self.obj_dict["sequence"] = None
 
             # Remove the compass point
@@ -760,10 +768,10 @@ class Node(Common):
 
         self.obj_dict["attributes"]["style"] = ",".join(styles)
 
-    def to_string(self):
+    def to_string(self, indent="", indent_level=1):
         """Return string representation of node in DOT language."""
-        # RMF: special case defaults for node, edge and graph properties.
-        #
+        indent_str = self.get_indent(indent, indent_level)
+
         node = quote_id_if_necessary(
             self.obj_dict["name"], unquoted_keywords=("graph", "node", "edge")
         )
@@ -775,7 +783,7 @@ class Node(Common):
         ):
             return ""
 
-        return f"{node}{self.attrs_string(prefix=' ')};"
+        return f"{indent_str}{node}{self.attrs_string(prefix=' ')};"
 
 
 __generate_attribute_methods(Node, NODE_ATTRIBUTES)
@@ -823,7 +831,6 @@ class Edge(Common):
             self.obj_dict["attributes"] = dict(attrs)
             self.obj_dict["type"] = "edge"
             self.obj_dict["parent_graph"] = None
-            self.obj_dict["parent_edge_list"] = None
             self.obj_dict["sequence"] = None
         else:
             self.obj_dict = obj_dict
@@ -905,13 +912,20 @@ class Edge(Common):
 
         return quote_id_if_necessary(node_str)
 
-    def to_string(self):
+    def to_string(self, indent="", indent_level=1):
         """Return string representation of edge in DOT language."""
         src = self.parse_node_ref(self.get_source())
         dst = self.parse_node_ref(self.get_destination())
 
+        indent_str = self.get_indent(indent, indent_level)
+
         if isinstance(src, frozendict):
-            edge = [Subgraph(obj_dict=src).to_string()]
+            sgraph = Subgraph(obj_dict=src)
+            edge = [
+                sgraph.to_string(
+                    indent=indent, indent_level=indent_level, inline=True
+                )
+            ]
         elif isinstance(src, int):
             edge = [str(src)]
         else:
@@ -923,13 +937,18 @@ class Edge(Common):
             edge.append("--")
 
         if isinstance(dst, frozendict):
-            edge.append(Subgraph(obj_dict=dst).to_string())
+            sgraph = Subgraph(obj_dict=dst)
+            edge.append(
+                sgraph.to_string(
+                    indent=indent, indent_level=indent_level, inline=True
+                )
+            )
         elif isinstance(dst, int):
             edge.append(str(dst))
         else:
             edge.append(dst)
 
-        return f"{' '.join(edge)}{self.attrs_string(prefix=' ')};"
+        return f"{indent_str}{' '.join(edge)}{self.attrs_string(prefix=' ')};"
 
 
 __generate_attribute_methods(Edge, EDGE_ATTRIBUTES)
@@ -1420,28 +1439,42 @@ class Graph(Common):
             for obj in obj_list:
                 Graph(obj_dict=obj).set_parent_graph(parent_graph)
 
-    def to_string(self):
+    def to_string(self, indent="", indent_level=0, inline=False):
         """Return string representation of graph in DOT language.
 
         @return: graph and subelements
         @rtype: `str`
         """
+        indent_str = self.get_indent(indent, indent_level)
+        child_indent = self.get_indent(indent, indent_level + 1)
+
         graph = []
 
-        if self.obj_dict.get("strict", None) is not None:
-            if self == self.get_parent_graph() and self.obj_dict["strict"]:
-                graph.append("strict ")
+        if not inline:
+            graph.append(indent_str)
+
+        first_line = []
+
+        if self == self.get_parent_graph() and self.obj_dict.get(
+            "strict", False
+        ):
+            first_line.append("strict")
 
         graph_type = self.obj_dict["type"]
-        if graph_type == "subgraph" and not self.obj_dict.get(
-            "show_keyword", True
-        ):
-            graph_type = ""
-        graph_name = quote_id_if_necessary(self.obj_dict["name"])
-        s = f"{graph_type} {graph_name} {{\n"
-        graph.append(s)
+        if graph_type != "subgraph" or self.obj_dict.get("show_keyword", True):
+            first_line.append(graph_type)
 
-        graph.extend(f"{a};\n" for a in self.formatted_attr_list())
+            # Suppressing the keyword hides the name as well
+            graph_name = self.obj_dict.get("name")
+            if graph_name:
+                first_line.append(quote_id_if_necessary(graph_name))
+
+        first_line.append("{\n")
+        graph.append(" ".join(first_line))
+
+        graph.extend(
+            f"{child_indent}{a};\n" for a in self.formatted_attr_list()
+        )
 
         edges_done = set()
 
@@ -1482,7 +1515,10 @@ class Graph(Common):
                     ):
                         continue
 
-                graph.append(node.to_string() + "\n")
+                node_str = node.to_string(
+                    indent=indent, indent_level=indent_level + 1
+                )
+                graph.append(f"{node_str}\n")
 
             elif obj["type"] == "edge":
                 edge = Edge(obj_dict=obj)
@@ -1490,14 +1526,21 @@ class Graph(Common):
                 if self.obj_dict.get("simplify", False) and edge in edges_done:
                     continue
 
-                graph.append(edge.to_string() + "\n")
+                edge_str = edge.to_string(
+                    indent=indent, indent_level=indent_level + 1
+                )
+                graph.append(f"{edge_str}\n")
                 edges_done.add(edge)
 
             else:
-                sgraph = Subgraph(obj_dict=obj)
-                graph.append(sgraph.to_string() + "\n")
+                sgraph_str = Subgraph(obj_dict=obj).to_string(
+                    indent=indent, indent_level=indent_level + 1
+                )
+                graph.append(f"{sgraph_str}\n")
 
-        graph.append("}\n")
+        graph.append(f"{indent_str}}}")
+        if not inline:
+            graph.append("\n")
 
         return "".join(graph)
 
@@ -1636,11 +1679,22 @@ class Dot(Graph):
         self.prog = "dot"
 
     def __getstate__(self):
-        dict = copy.copy(self.obj_dict)
-        return dict
+        state = {
+            "obj_dict": copy.copy(self.obj_dict),
+            "prog": self.prog,
+            "shape_files": copy.deepcopy(self.shape_files),
+            "formats": copy.copy(self.formats),
+        }
+        return state
 
     def __setstate__(self, state):
-        self.obj_dict = state
+        if "obj_dict" not in state:
+            # Backwards compatibility for old picklings
+            state = {"obj_dict": state}
+        self.obj_dict = state.get("obj_dict", {})
+        self.prog = state.get("prog", "dot")
+        self.shape_files = state.get("shape_files", [])
+        self.formats = state.get("formats", OUTPUT_FORMATS)
 
     def set_shape_files(self, file_paths):
         """Add the paths of the required image files.
@@ -1776,44 +1830,36 @@ class Dot(Graph):
             args = []
 
         # temp file
-        tmp_fd, tmp_name = tempfile.mkstemp()
-        os.close(tmp_fd)
-        self.write(tmp_name, encoding=encoding)
-        tmp_dir = os.path.dirname(tmp_name)
+        with tempfile.TemporaryDirectory(
+            ignore_cleanup_errors=True
+        ) as tmp_dir:
+            fp = tempfile.NamedTemporaryFile(dir=tmp_dir, delete=False)
+            fp.close()
+            self.write(fp.name, encoding=encoding)
 
-        # For each of the image files...
-        for img in self.shape_files:
-            # Get its data
-            f = open(img, "rb")
-            f_data = f.read()
-            f.close()
-            # And copy it under a file with the same name in
-            # the temporary directory
-            f = open(os.path.join(tmp_dir, os.path.basename(img)), "wb")
-            f.write(f_data)
-            f.close()
+            # For each of the image files, copy it to the temporary directory
+            # with the same filename as the original
+            for img in self.shape_files:
+                outfile = os.path.join(tmp_dir, os.path.basename(img))
+                with open(img, "rb") as img_in, open(outfile, "wb") as img_out:
+                    img_data = img_in.read()
+                    img_out.write(img_data)
 
-        arguments = [f"-T{format}"] + args + [tmp_name]
+            arguments = [f"-T{format}"] + args + [fp.name]
 
-        try:
-            stdout_data, stderr_data, process = call_graphviz(
-                program=prog,
-                arguments=arguments,
-                working_dir=tmp_dir,
-            )
-        except OSError as e:
-            if e.errno == errno.ENOENT:
-                args = list(e.args)
-                args[1] = f'"{prog}" not found in path.'
-                raise OSError(*args)
-            else:
-                raise
-
-        # clean file litter
-        for img in self.shape_files:
-            os.unlink(os.path.join(tmp_dir, os.path.basename(img)))
-
-        os.unlink(tmp_name)
+            try:
+                stdout_data, stderr_data, process = call_graphviz(
+                    program=prog,
+                    arguments=arguments,
+                    working_dir=tmp_dir,
+                )
+            except OSError as e:
+                if e.errno == errno.ENOENT:
+                    args = list(e.args)
+                    args[1] = f'"{prog}" not found in path.'
+                    raise OSError(*args)
+                else:
+                    raise
 
         if process.returncode != 0:
             code = process.returncode
